/* This file is part of the unreleased c-torture package.
   Copyright (C) 1995 Free Software Foundation, Inc.  */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/resource.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

extern char **environ;

char *default_passes[] =
{
  "",
  "-O -fno-omit-frame-pointer",
  "-O -fomit-frame-pointer",
  "-O2 -fno-omit-frame-pointer",
  "-O2 -fomit-frame-pointer",
  "-O3 -fno-omit-frame-pointer",
  "-O3 -fomit-frame-pointer",
  "-O3 -funroll-all-loops",
  NULL
};

char *run_cmd = NULL;
char *global_extra_args = NULL;
char *global_post_args = NULL;
char **passes = NULL;
int n_passes = 0;
int flag_verbose = 0;
int flag_execute = 1;
int failure_count;
int compiler_timeout = 0;
int runtime_timeout = 0;

char *prog_name;
char *tmpdir;
char *tmpname;

void run_a_test (char *, char *, char *[], char *, int, int);
char *readfile (char *);
char *makepath (char *, char *);
void cleanup_and_exit (int);
void no_core_dumps (void);
void set_watchdog (int);
int split_string_into_args (char **, int, char *);
void fatal (const char *, ...);
void usage (void);

main (int argc, char *argv[])
{
  char *compiler_args[1000];	/* ??? static alloc ??? */
  int i, ii, j;
  int argbase, arglim, argi;
  struct stat stat_buf;
  int sysr;

  prog_name = argv[0];
  passes = malloc (1);

  argv++;
  argc--;

  while (argc > 0)
    {
      char *arg = argv[0];
      int arglen;

      arg += (arg[0] == '-' && arg[1] == '-');	/* allow --cmd-line-option */
      arglen = strlen (arg);
      if (arglen < 2)
	arglen = 2;

      if (!strncmp (arg, "-help", arglen))
	{
	  usage ();
	  exit (0);
	}
      if (!strncmp (arg, "-verbose", arglen))
	flag_verbose += 1;
      else if (!strncmp (arg, "-no-run", arglen))
	flag_execute = 0;
      else if (!strncmp (arg, "-run", arglen))
	{
	  if (argc < 2)
	    fatal ("argument expected after %s", argv[0]);
	  argv++;
	  argc--;
	  run_cmd = argv[0];
	  if (!runtime_timeout)
	    runtime_timeout = 1000;	 /* bump up timeout if running under a simulator */
	}
      else if (!strncmp (arg, "-options", arglen))
	{
	  if (argc < 2)
	    fatal ("argument expected after %s", argv[0]);
	  argv++;
	  argc--;
	  if (global_extra_args != NULL)
	    fatal ("cannot handle multiple -options arguments");
	  global_extra_args = argv[0];
	}
      else if (!strncmp (arg, "-postopts", arglen))
	{
	  if (argc < 2)
	    fatal ("argument expected after %s", argv[0]);
	  argv++;
	  argc--;
	  if (global_post_args != NULL)
	    fatal ("cannot handle multiple -postopts arguments");
	  global_post_args = argv[0];
	}
      else if (!strncmp (arg, "-extra-pass", arglen))
	{
	  if (argc < 2)
	    fatal ("argument expected after %s", argv[0]);
	  argv++;
	  argc--;
	  passes = realloc (passes, (n_passes + 1) * sizeof (char *));
	  passes[n_passes++] = argv[0];
	}
      else if (!strncmp (arg, "-no-builtin-passes", arglen))
	{
	  default_passes[0] = NULL;
	}
      else if (!strncmp (arg, "-cc-timeout", arglen))
	{
	  if (argc < 2)
	    fatal ("argument expected after %s", argv[0]);
	  argv++;
	  argc--;
	  compiler_timeout = (int) strtol (*argv, (char **)0, 0);
	}
      else if (!strncmp (arg, "-run-timeout", arglen))
	{
	  if (argc < 2)
	    fatal ("argument expected after %s", argv[0]);
	  argv++;
	  argc--;
	  runtime_timeout = (int) strtol (*argv, (char **)0, 0);
	}
      else
	break;

      argv++;
      argc--;
    }

  if (!compiler_timeout)
    compiler_timeout = 100;

  if (!runtime_timeout)
    runtime_timeout = 100;

  if (argc == 0)
    fatal ("compiler spec expected");

  sysr = stat (argv[0], &stat_buf);
  if (sysr == -1)
    fatal ("%s: no such compiler or compiler directory", argv[0]);
  if ((stat_buf.st_mode & S_IFMT) == S_IFDIR)
    {
      compiler_args[0] = malloc (strlen (argv[0]) + 6);
      strcpy (compiler_args[0], argv[0]);
      strcat (compiler_args[0], "/xgcc");
      compiler_args[1] = malloc (strlen (argv[0]) + 4);
      strcpy (compiler_args[1], "-B");
      strcat (compiler_args[1], argv[0]);
      strcat (compiler_args[1], "/");
      argbase = 2;
    }
  else
    {
      compiler_args[0] = argv[0];
      argbase = 1;
    }

  argv++;
  argc--;

  tmpdir = getenv ("TMPDIR");
  if (tmpdir == NULL)
    tmpdir = "/tmp";
  tmpname = makepath (tmpdir, "ctXXXXXX");
  mktemp (tmpname);

  if (signal (SIGINT, SIG_IGN) != SIG_IGN)
    signal (SIGINT, cleanup_and_exit);
#ifdef SIGHUP
  if (signal (SIGHUP, SIG_IGN) != SIG_IGN)
    signal (SIGHUP, cleanup_and_exit);
#endif
  if (signal (SIGTERM, SIG_IGN) != SIG_IGN)
    signal (SIGTERM, cleanup_and_exit);

  /* Add the list of builtin compiler options to use for the passes.  */
  for (i = 0; default_passes[i] != NULL; i++)
    {
      passes = realloc (passes, (n_passes + 1) * sizeof (char *));
      passes[n_passes++] = default_passes[i];
    }

  for (i = 0; i < argc; i++)
    {
      FILE *dirp;
      char filename[1024];
      char *descfile;
      char *buf;
      int execute_these_files = 0;
      int expect_failure = 0;

      descfile = makepath (argv[i], "DESC");
      buf = readfile (descfile);
      if (buf != NULL)
	{
	  if (strncmp ("compile-pass", buf, 12) == 0)
	    {
	      execute_these_files = 0;
	      expect_failure = 0;
	    }
	  else if (strncmp ("compile-fail", buf, 12) == 0)
	    {
	      execute_these_files = 0;
	      expect_failure = 1;
	    }
	  else if (strncmp ("execute-pass", buf, 12) == 0)
	    {
	      execute_these_files = flag_execute;
	      expect_failure = 0;
	    }
	  else
	    fatal ("wrong syntax in DESC file");
	  free (buf);
	}
      else
	printf ("warning: no DESC file in tests directory\n");
      free (descfile);

      sprintf (filename, "cd %s; ls -r *.c", argv[i]);
      dirp = popen (filename, "r");
      if (dirp == NULL)
	fatal ("cannot open test directory");

      printf ("TEST DIR: %s\n", argv[i]);
      fflush (stdout);
      failure_count = 0;

      for (;;)
	{
	  int len;
	  char *test_name, *stripped_test_name;

	  if (fscanf (dirp, "%s", filename) < 1)
	    break;
	  len = strlen (filename);
	  if (len <= 2 || strcmp (".c", filename + len - 2) != 0)
	    continue;
	  stripped_test_name = filename;
	  test_name = makepath (argv[i], stripped_test_name);

	  printf ("%s\n", stripped_test_name);
	  fflush (stdout);

	  argi = argbase;

	  for (ii = 0; ii < n_passes; ii++)
	    {
	      char all_args[200];

	      sprintf (all_args, "%s %s", passes[ii],
		       global_extra_args ? global_extra_args : "");

	      argi = argbase;
	      argi = split_string_into_args (compiler_args, argi, passes[ii]);

	      if (global_extra_args != NULL)
		argi = split_string_into_args (compiler_args, argi,
					       global_extra_args);
	      arglim = argi;

	      if (! execute_these_files)
		compiler_args[argi++] = "-S";

	      compiler_args[argi++] = "-w";
	      compiler_args[argi++] = "-o";
	      compiler_args[argi++] = tmpname;
	      compiler_args[argi++] = test_name;

	      if (global_post_args != NULL)
		argi = split_string_into_args (compiler_args, argi,
					       global_post_args);

	      compiler_args[argi++] = NULL;

	      if (flag_verbose)
		{
		  printf ("%s %s\n", test_name, all_args);
		  fflush (stdout);
		}

	      if (flag_verbose > 1)
		{
		  char **argp;
		  printf ("+");
		  for (argp = compiler_args; *argp != NULL; argp++)
		    printf(" %s", *argp);
		  printf ("\n");
		}

	      run_a_test (compiler_args[0],
			  all_args,
			  compiler_args,
			  stripped_test_name,
			  execute_these_files,
			  expect_failure);

	      for (j = argbase; j < arglim; j++)
		free (compiler_args[j]);
	    }

	  free (test_name);
	}

      pclose (dirp);

      if (failure_count == 0)
	printf ("Test completed successfully\n");
      else
	printf ("Test completed with %d failure%s\n",
		failure_count, "s" + (failure_count == 1));
      fflush (stdout);
    }

  unlink (tmpname);
  exit (0);
}

void
run_a_test (char *compiler_path,
	    char *compiler_args_as_string,
	    char *compiler_args[],
	    char *test_name,
	    int execute_these_files,
	    int expect_failure)
{
  pid_t pid, wait_pid;
  int errmess[2];
  int error_flag = 0;
  int sysr;

  sysr = pipe (errmess);
  if (sysr == -1)
    fatal ("pipe call failed");

  /* flush any buffered output before starting run */
  fflush(stdout);

  pid = fork ();
  if (pid == -1)
    fatal ("fork call failed");

  if (pid != 0)
    {
      int n_total_read = 0;
      int status;
      char buf[8000];

      close (errmess[1]);

      /* Read error output from compiler.  If there is nothing to read,
	 the read call will fail when the child terminates.  */
      for (;;)
	{
	  int n_read;
	  n_read = read (errmess[0], buf + n_total_read, 7999 - n_total_read);
	  if (n_read <= 0)
	    break;
	  n_total_read += n_read;
	}

      buf[n_total_read] = '\0';

      /* log the output, hope there was a \n at the end */
      if (flag_verbose > 1)
	fwrite (buf, strlen (buf), sizeof (char), stdout);

      do {
	wait_pid = wait (&status);
      } while (wait_pid != -1 && wait_pid != pid);
      
      if (WIFEXITED (status) && WEXITSTATUS (status) != 0)
	{
	  char *where;

	  if (n_total_read == 0)
	    printf ("ERROR: %s: compiler returned error status w/o error message (asm error?): %s\n",
		    test_name, compiler_args_as_string);
	  else
	    {
	      where = strstr (buf, "fatal signal");
	      if (where != NULL)
		{
		  int signum;
		  sscanf (where + strlen ("fatal signal") + 1, "%d", &signum);
		  if (signum == SIGALRM || signum == SIGVTALRM)
		    {
		      printf ("ERROR: %s: compiler seems to hang: %s\n",
			      test_name, compiler_args_as_string);
		      failure_count++;
		      error_flag = 3;
		      fflush (stdout);
		    }
		  else
		    {
		      printf ("ERROR: %s: compiler got signal %d: %s\n",
			      test_name, signum, compiler_args_as_string);
		      failure_count++;
		      error_flag = 2;
		      fflush (stdout);
		    }
		}
	      else
		{
		  if (! expect_failure)
		    {
		      printf ("ERROR: %s: compiler returns exit status %d: %s\n",
			      test_name, WEXITSTATUS (status),
			      compiler_args_as_string);
		      failure_count++;
		      fflush (stdout);
		    }
		  error_flag = 1;
		}
	    }
	}
      else if (WIFSIGNALED (status))
	{
	  printf ("ERROR: %s: compiler got signal %d: %s\n",
		  test_name, WTERMSIG (status), compiler_args_as_string);
	  failure_count++;
	  error_flag = 2;
	  fflush (stdout);
	}

      if (expect_failure && error_flag == 0)
	{
	  printf ("ERROR: %s: compiler accepts syntax error: %s\n",
		  test_name, compiler_args_as_string);
	  failure_count++;
	  error_flag = 4;
	  fflush (stdout);
	}

      close (errmess[0]);
    }
  else
    {
      /* Make compiler's error output go to the `errmess' pipe.  */
      close (errmess[0]);
      dup2 (errmess[1], fileno (stderr));
      close (errmess[1]);

      set_watchdog (compiler_timeout);
      no_core_dumps ();
      execve (compiler_path, compiler_args, environ);

      fatal ("cannot execute compiler: %s", compiler_path);
    }

  if (error_flag || ! execute_these_files)
    return;

  pid = fork ();
  if (pid != 0)
    {
      int status;

      do {
	wait_pid = wait (&status);
      } while (wait_pid != -1 && wait_pid != pid);

      if (WIFEXITED (status) && WEXITSTATUS (status) != 0)
	{
	  printf ("ERROR: %s: compiled program returns exit status %d: %s\n",
		  test_name, WEXITSTATUS (status), compiler_args_as_string);
	  failure_count++;
	  fflush (stdout);
	}
      else if (WIFSIGNALED (status))
	{
	  printf ("ERROR: %s: compiled program got signal %d: %s\n",
		  test_name, WTERMSIG (status), compiler_args_as_string);
	  failure_count++;
	  fflush (stdout);
	}
    }
  else
    {
      int save_stdout = dup (fileno (stdout));

      /* Ignore test's error output (unless very verbose).  */
      if (flag_verbose < 2)
	{
	  freopen ("/dev/null", "w", stderr);
	  freopen ("/dev/null", "w", stdout);
	}

      set_watchdog (runtime_timeout);
      no_core_dumps ();
      if (run_cmd != NULL)
	{
	  char *args[100];
	  int argi = 0;

	  argi = split_string_into_args (args, argi, run_cmd);
	  args[argi++] = tmpname;
	  args[argi++] = NULL;

	  if (argi >= (sizeof(args) / sizeof(*args))) {
	    if (flag_verbose < 2)
	      dup2 (save_stdout, fileno (stdout));
	    fatal ("run: args vector overflow\n");
	  }
	  if (flag_verbose >= 2)
	    {
	      char **arg;
	      printf ("+");
	      for (arg = args; *arg != NULL; arg++)
		printf (" %s", *arg);
	      printf ("\n");
	    }
	  execve (args[0], args, environ);
	}
      else
	{
	  char *args[2];

	  args[0] = tmpname;
	  args[1] = NULL;
	  execve (tmpname, args, environ);
	}

      if (flag_verbose < 2)
	dup2 (save_stdout, fileno (stdout));
      fatal ("cannot execute test: %s", test_name);
    }
}

int
split_string_into_args (char **argslist,
			int argi,
			char *command_string)
{
  char *outp, *x;
  int arg_len;

  outp = x = command_string;
  while (*x != '\0')
    {
      x = strchr (outp, ' ');
      if (x == NULL)
	x = outp + strlen (outp);
      arg_len = x - outp;
      argslist[argi] = malloc (arg_len + 1);
      memcpy (argslist[argi], outp, arg_len);
      argslist[argi][arg_len] = '\0';
      outp = x + 1;
      argi++;
    }
  return argi;
}

/* Damn H-PUX handles RLIMIT_CORE, but the definition of RLIMIT_CORE
   is hidden within #ifdef _KERNEL.  */
#ifndef RLIMIT_CORE
#define RLIMIT_CORE 4
#endif

void
no_core_dumps (void)
{
  struct rlimit rlp;
  rlp.rlim_cur = 0;
  rlp.rlim_max = 0;
  setrlimit (RLIMIT_CORE, &rlp);
}

void
set_watchdog (int seconds)
{
  int sysr;
  struct itimerval itimer;

  itimer.it_interval.tv_sec = 0;
  itimer.it_interval.tv_usec = 0;
  itimer.it_value.tv_sec = seconds;
  itimer.it_value.tv_usec = 0;
  sysr = setitimer (1, &itimer, NULL);
  if (sysr != 0)
    fatal ("setitimer call failed");
}

char *
makepath (char *dir, char *file)
{
  int ldir = strlen (dir);
  int lfile = strlen (file);
  char *ret;

  ret = malloc (ldir + lfile + 2);
  ldir -= (dir[ldir - 1] == '/');
  memcpy (ret, dir, ldir);
  ret[ldir++] = '/';
  memcpy (ret + ldir, file, lfile);
  ret[ldir + lfile] = '\0';
  return ret;
}

char *
readfile (char *path)
{
  int fd;
  struct stat sb;
  char *buf;
  fd = open (path, O_RDONLY, 0);
  if (fd == -1)
    return NULL;
  fstat (fd, &sb);
  buf = malloc (sb.st_size);
  read (fd, buf, sb.st_size);
  return buf;
}

void
cleanup_and_exit (int signum)
{
  unlink (tmpname);
  signal (signum, SIG_DFL);
  kill (getpid (), signum);
}

void
fatal (const char *ctrl, ...)
{
  va_list args;

  va_start (args, ctrl);
  vprintf (ctrl, args);
  printf ("\n");
  va_end (args);
  exit (1);
}

void
usage (void)
{
  printf ("%s [OPTIONS] compiler-spec test-dir [test-dir ... ]\n", prog_name);
  puts ("Where OPTIONS is any sensible combination from this list:");
  puts ("  [ -options \"arg1 arg2...\" ] - pass args to every compiler invocation");
  puts ("                                  example: -options \"-I/foo/include -fpic\"");
  puts ("                                  running: gcc -I/foo/include -fpic <test>.c");
  puts ("  [ -postopts \"...argN-1 argN\" ] - append args to every compiler invocation");
  puts ("                                     example: -postopts \"-L/my/lib -lc\"");
  puts ("                                     running: gcc <test>.c -L/my/lib -lc");
  puts ("  [ -extra-pass \"arg\" ]     - make a test pass with arg passed to compiler");
  puts ("                                example: -extra-pass \"-O9 -fno-force-mem\"");
  puts ("  [ -verbose ]              - be verbose...");
  puts ("  [ -no-run ]               - just compile, don't execute");
  puts ("  [ -run \"run_cmd\" ]        - when executing test, let run_cmd run it");
  puts ("  [ -no-builtin-passes ]    - only use the passes spec'd by -extra-pass");
  puts ("  [ -cc-timeout \"secs\" ]    - change timeout for running compiler");
  puts ("  [ -run-timeout \"secs\" ]   - change timeout for running test");
  puts ("compiler-spec               - path to compiler or compiler directory");
  puts ("test-dir [test-dir ... ]    - c-torture test directory paths (execute, ...)");
  puts ("(Unambiguous abbreviations of options are permitted)");
}
